import pandas as pd
from flex_dtk import trading_db


def test_query_returns_result_of_pandas_read_sql(monkeypatch):
    def dummy_df(*args, **kwargs) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [1, 2, 3],
            }
        )

    monkeypatch.setattr(trading_db.pd, "read_sql", dummy_df)
    actual = trading_db.query(
        "SELECT * FROM TABLE",
        server="server.cool",
        database="db",
        username="m3",
        password="s3cr3t!123",
    )
    assert set(actual.columns) == {"a", "b"}
    assert len(actual) == 3



"""
The `interactive` module will run an interactive duckdb session between the various
data resources.

Note that just importing this module, if in an IPython environment, will register the
magic `%%sql` command to your session.

SQL queries runnable in [duckdb](https://duckdb.org/docs/sql/introduction) which is
designed to closely follow the structure of Postgres Queries.

The interactive SQL session will be kept in memory at run-time *not ran on a SQL database*.
It's worthwhile therefore, to only bring in the data you need before querying it, in order
to avoid overloading your memory.

For example:

```python
from flex_dtk import interactive

interactive.register_warehouse_query(
    "big_table",
    "SELECT * FROM a_big.table",
    server,
    database,
)

df = interactive.flexsql("SELECT * FROM big_table LIMIT 10")
```

Will load *everything* from `a_big.table` into memory before filtering it, in comparison to
the below which will only load 10 items into memory to start with:

```python
from flex_dtk import interactive

interactive.register_warehouse_query(
    "small_table",
    "SELECT TOP 10 * FROM a_big.table",
    server,
    database,
)

df = interactive.flexsql("SELECT * FROM small_table")
```
"""

from datetime import datetime

try:
    import duckdb
    import pandas as pd
    from IPython.core.magic import register_cell_magic
except ImportError as exception:
    raise ImportError(
        "You may be missing required dependencies for flex_dtk's interactive module\n"
        "Hint: try installing 'flex_dtk[interactive]' rather than simply 'flex_dtk'"
    ) from exception

from flex_dtk import datalake, warehouse, trading_db

pd.set_option("display.max_columns", None)
pd.set_option("display.max_rows", 100)

connection = duckdb.connect()


def flexsql(query: str) -> pd.DataFrame:
    """
    Execute a SQL query in the current session, and return the results as a
    pandas dataframe.
    """
    return connection.execute(query).fetchdf()


def register_flexsql(table: str, query: str) -> None:
    """
    Execute a SQL query in the current session, and register the result to a
    virtual table.
    """
    connection.register(table, flexsql(query))
    print(f"Result registered to {table}")


try:

    @register_cell_magic
    def sql(line, cell):  # pragma: no cover
        """
        Jupyter notebook magic implementation of lakesql function.

        Will  execute contents of cell as SQL query, if line is given,
        they'll also be registered to that line as a virtual table.
        """
        dataframe = flexsql(cell)
        if line:
            connection.register(line, dataframe)
            print(f"result registered to {line}")
        return flexsql(cell)
except NameError:
    pass


def _ensure_datetime(target: str | datetime) -> datetime:
    """
    Internal function, if target is string, will convert
    to a datetime under format "%Y-%m-%d" (or raise error)
    otherwise, will pass back target.
    """
    if isinstance(target, str):
        return datetime.strptime(target, "%Y-%m-%d")
    return target


def _registerable_name_from_path(path: str) -> str:
    """
    Internal function, decide name to register table against from path.
    """
    return path.split("/", maxsplit=2)[-1].replace("/", "_")


def register_datalake_date_range(
    path: str,
    from_date: str | datetime,
    to_date: str | datetime,
    **kwargs,
) -> None:
    """
    Read date range of datalake partitions, and register them as a
    virtual table.
    """
    from_date = _ensure_datetime(from_date)
    to_date = _ensure_datetime(to_date)
    connection.register(
        (register_name := _registerable_name_from_path(path)),
        datalake.between_dates(
            path,
            from_date=from_date,
            to_date=to_date,
            **kwargs,
        ),
    )
    print(f"Table registered to {register_name}")


def register_datalake_latest(
    path: str,
    as_of: str | datetime | None = None,
    **kwargs,
) -> None:
    """
    Read latest datalake partition, and register as a
    virtual table.
    """
    as_of = as_of if as_of is None else _ensure_datetime(as_of)
    connection.register(
        (register_name := _registerable_name_from_path(path)),
        datalake.latest(
            path,
            as_of=as_of,
            **kwargs,
        ),
    )
    print(f"Table registered to {register_name}")


def register_datalake_for_date(
    path: str,
    for_date: str | datetime,
    **kwargs,
) -> None:
    """
    Read a single datalake partition, and register as a
    virtual table.
    """
    for_date = _ensure_datetime(for_date)
    connection.register(
        (register_name := _registerable_name_from_path(path)),
        datalake.for_date(
            path=path,
            for_date=for_date,
            **kwargs,
        ),
    )
    print(f"Table registered to {register_name}")


def register_warehouse_query(
    register_name: str,
    query: str,
    server: str,
    database: str,
    **kwargs,
) -> None:
    """
    Query the warehouse and register results as an in-session virtual table.
    """
    connection.register(
        register_name,
        warehouse.query(
            query,
            server=server,
            database=database,
            **kwargs,
        ),
    )
    print(f"Table registered to {register_name}")


def register_trading_db_query(
    register_name: str,
    query: str,
    server: str,
    database: str,
    username: str,
    password: str,
    **kwargs,
) -> None:
    """
    Query the trading database and register results as an in-session virtual table
    """
    connection.register(
        register_name,
        trading_db.query(
            query,
            server=server,
            database=database,
            username=username,
            password=password,
            **kwargs,
        ),
    )
    print(f"Table registered to {register_name}")



from datetime import datetime

import pandas as pd

from flex_dtk import datalake


class DummyFileSystem:
    def __init__(*args, **kwargs) -> None: ...

    def clear_instance_cache(*args, **kwargs) -> None: ...

    def ls(*args, **kwargs) -> list[str]:
        return ["file/2024", "file/2022"]

    def glob(self, *args, **kwargs) -> list[str]:
        return self.ls()


def return_args(*args, **kwargs) -> dict:
    return {
        "args": args,
        "kwargs": kwargs,
    }


def test_data_functions_call_dataframe_from_files_with_expected_arguments(
    monkeypatch,
) -> None:
    monkeypatch.setattr(datalake.fsspec, "filesystem", DummyFileSystem)
    monkeypatch.setattr(
        datalake, "_latest_partition_folder", lambda x, y: "cool/folder"
    )
    monkeypatch.setattr(datalake, "_dataframe_from_files", return_args)
    actual = datalake.latest("some/file/somewhere/or/other")
    for to_call, kwargs in [
        (datalake.latest, {"path": "some/file/somewhere/or/other"}),
        (
            datalake.between_dates,
            {
                "path": "some/file/somewhere/or/other",
                "from_date": datetime(2023, 2, 2),
                "to_date": datetime(2023, 4, 5),
            },
        ),
        (
            datalake.for_dates,
            {
                "path": "some/file/somewhere/or/other",
                "dates": [datetime(2023, 4, 1), datetime(2024, 2, 3)],
            },
        ),
        (
            datalake.for_date,
            {
                "path": "some/file/somewhere/or/other",
                "target_date": datetime(2024, 2, 3),
            },
        ),
    ]:
        actual = to_call(**kwargs)
        assert actual["kwargs"]["storage_options"] == {
            "account_name": "some",
            "anon": False,
            "use_listings_cache": False,
        }
        assert set(actual["args"][0]) == {
            "file/2024",
            "file/2022",
        }


def test_latest_partition_folder_returns_expected(monkeypatch) -> None:
    monkeypatch.setattr(datalake.fsspec, "filesystem", DummyFileSystem)
    actual = datalake._latest_partition_folder(
        "some/really/cool/path/2023", datetime(2024, 3, 1)
    )
    assert actual == "file/2024"


def test_all_files_in_path_returns_expected(monkeypatch) -> None:
    actual = datalake._all_files_in_path(
        "cool/path",
        DummyFileSystem(),
    )
    assert set(actual) == {"file/2024", "file/2022"}


def test_dataframe_from_files_concatenates_job_results(monkeypatch) -> None:
    def dummy_file(*args, **kwargs) -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2], "b": [1, 2]})

    actual = datalake._dataframe_from_files(
        ["one", "two", "three"], {}, input_operation=dummy_file
    )
    assert set(actual.columns) == {"a", "b"}
    assert len(actual) == 6



"""
Functions for reading SQL queries from the azure SQL warehouse.

All functions will authenticate using Azure Default Credentials. If you are
are working locally the easiest way to set this up is to raise a ticket for
Azure CLI to be installed and set up on your machine. For more complex cases
like authenticating on a remote application, see Azure's own
[docs](https://learn.microsoft.com/en-us/dotnet/api/azure.identity.defaultazurecredential)
"""

import struct
import warnings
import pyodbc
from sqlalchemy.engine import Engine

import pandas as pd

from azure.identity import DefaultAzureCredential


def azure_connection(
    server: str, database: str
) -> pyodbc.Connection:  # pragma: no cover
    """
    Build an azure connection with Default Credential authentication
    """
    possible_drivers: list = [
        i for i in pyodbc.drivers() if "SQL Server" in i and i != "SQL Server"
    ]
    if len(possible_drivers) == 0:
        msg = (
            "No compatible MSSQL driver could be found on this device.\n"
            "(note that 'SQL Server' itself does not support active directory "
            "connection)\n\n"
            "Consider raising a ticket to have SQL Server 17, 18 or later installed."
        )
        raise RuntimeError(msg)
    driver: str = possible_drivers[0]
    connection_string: str = (
        f"Driver={{{driver}}};Server=tcp:{server},1433;Database={database};"
        "Encrypt=yes;TrustServerCertificate=No;Connection Timeout=30"
    )
    credential = DefaultAzureCredential()
    token_bytes = credential.get_token(
        "https://database.windows.net/.default"
    ).token.encode("UTF-16-LE")
    token_struct = struct.pack(f"<I{len(token_bytes)}s", len(token_bytes), token_bytes)
    connection = pyodbc.connect(connection_string, attrs_before={1256: token_struct})
    return connection


def query(
    sql: str,
    server: str | None = None,
    database: str | None = None,
    connection: pyodbc.Connection | Engine | str | None = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Run and return the results of a database query for the Azure SQL Warehouse.

    Additional keyword arguments will be passed down to pandas' `read_sql`.

    Note, unless other keyword arguments are passed down, a pyarrow dtype_backend will
    be used by default.

    Connection is an optional parameter to allow for the use of a
    pre-existing connection so doing batch queries is more efficient.
    Use connection with context handler to ensure connection is closed.
    """
    if not ((server and database) or connection):
        raise ValueError("Either sever and database, "
        "or existing connection must be given to connect to SQL")

    kwargs = kwargs or {"dtype_backend": "pyarrow"}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return pd.read_sql(
            sql,
            con = connection or azure_connection(str(server), str(database)),
            **kwargs,
        )



"""
Flexitricity data toolkit is a set of utility functions for loading data, and
optionally, also for interactive with data sources via SQL.

For distribution reasons `flex_dtk` is open source, but it's unlikely to be of
actual value outside the organisation.

To get up and running with `flex_dtk` you can pip install the package with:

```bash
pip install "flex_dtk[interactive]"
```

or, if you don't want to include the interactive (jupyter notebook, duckdb)
dependencies, simply:

```bash
pip install flex_dtk
```

For all functionality to work, you'll need some additional set up too:

- Azure credentials set up on your machine, this is easiest done by raising
 ticket for Azure CLI to be installed and set up on your machine.

- An MSSQL driver installed (most likely 17 or 18) in order to access the Azure
SQL database.

(You don't need a postgres driver, since `flex_dtk` uses precompiled binaries
from the `pyscopg2-binary` library).

`flex_dtk` is organised into four modules:
- datalake (functions for loading data from the datalake)
- warehouse (functions for loading data from the SQL warehouse)
- trading_db (functions for loading data from the trading database)
- interactive (optional module for interacting with the above data-sources using SQL)
"""

from flex_dtk._version import __version__  # noqa
